import random

# seek t such that D[t - 1] <= r < D[t]
# dichotomic search
def seekCDist(D, r):
    t_min = min(D)
    t_max = max(D)
    if r < D[t_min]:
        return t_min
    while t_max - t_min > 1:
        t = int((t_min + t_max) / 2)
        if r < D[t]:
            t_max = t
        else:
            t_min = t
    return t_max

# pick a value according to a (cumulative) distribution D. The
# distribution is not normalized, each D[i] is a positive integer and
# D[i] grows with i. The actual probability to pick i is D[i]/max(D[i])
def pickFromCDist(D):
    M = max(D[i] for i in D)
    r = random.randrange(0, M)
    return seekCDist(D, r)

##############
# DecodeV
##############

def acceptV(weight, rV):
    if weight in rV:
        return random.random() <= rV[weight]
    else:
        return False

def decodeV(sV, sk, DV, rV, delta):
    F = sk.HUt.base_ring()
    n2 = sk.HVt.nrows()
    kV = n2 - sk.HVt.ncols()
    support = range(n2)
    random.shuffle(support)
    # We add sV as the (n2+1)-th row of HVt. This will transform the
    # syndrome as needed without extra operation. This require a few
    # extra as [n2] at the end of the support, and later [-1] at the
    # end of x
    support += [n2]
    Ht, support = echelon_form_tr(sk.HVt.stack(sV), support)
    # support has possibly been reordered to have the pivot positions first
    x = vector(F, [0] * n2 + [-1]) # x will have -1 in a last (n2+1)-th coordinate
    while True:
        t = pickFromCDist(DV)
        # first n2 - kV positions of x[support[]] are and remain zero
        for i in range(n2 - kV, n2 - kV + delta):
            x[support[i]] = random.choice(F.list())
        for i in range(n2 - kV + delta, n2 - t):
            x[support[i]] = 0
        for i in range(n2 - t, n2):
            x[support[i]] = random.choice(F.list_of_elements_of_multiplicative_group())
        y = x * Ht
        if acceptV(x.hamming_weight() + y.hamming_weight(), rV): # rejection sampling
            break
    for i in range(n2 - kV):
        x[support[i]] = -y[i]
    eV = vector(F, x[:-1]) # ignore last coordinate of x
    return eV

##############
# DecodeU
##############

# return a reordered version of of the integers [0, len(eV)] such that
# the last k_nz element correspond to non-zero coordinates of eV. The
# other non-zero coordinates of eV are placed at the beginning.  This
# is done to insure that (with overwhelming probability) when we take
# the last kU positions (kU >= k_nz) exactly k_nz of them are non-zero
# coordinates of eV.
def randomSupportSpecialU(k_nz, eV):
    n2 = len(eV)
    NZpos = eV.support()
    Zpos = list(set(range(n2)).difference(NZpos))
    random.shuffle(NZpos)
    random.shuffle(Zpos)
    l = len(NZpos) - k_nz
    return NZpos[:l] + Zpos + NZpos[l:]
def ellUV(e):
    n2 = int(len(e) / 2)
    l = 0
    for i in range(n2):
        if ((e[i] * e[i + n2] == 0) and (e[i] + e[i + n2] != 0)):
            l += 1
    return l
def acceptU(t, l, rU):
    return (random.random() <= rU[t][l])

def decodeU(sU, sk, eV, DU, rU, w, delta):
    F = sk.HUt.base_ring()
    n2 = sk.HUt.nrows()
    kU = n2 - sk.HUt.ncols()
    t = eV.hamming_weight()
    while True:
        k_nz = pickFromCDist(DU[t])
        k_z = kU - delta - k_nz
        # support verifies:
        # 1- last k_nz positions of support are in supp(eV)
        # 2- the other positions in supp(eV) come first in support
        support = randomSupportSpecialU(k_nz, eV)
        # We add sU as the (n2+1)-th row of HUt. See comment in decodeV.
        support += [n2]
        Ht, support = echelon_form_tr(sk.HUt.stack(sU), support)
        # support has possibly been reordered to have the pivot positions first
        # but conditions 1- and 2- above are still met
        # => the information set support[n2-kU:n2] intersect supp(eV) in exactly k_nz positions
        x = vector(F, [0] * n2 + [-1]) # x will have -1 in a last (n2+1)-th coordinate
        while True:
            # first n2 - kU positions of x[support[]] are and remain zero
            for i in range(n2 - kU, n2 - kU + delta):
                x[support[i]] = random.choice(F.list())
            for i in range(n2 - kU + delta, n2 - k_nz):
                x[support[i]] = random.choice(F.list_of_elements_of_multiplicative_group())
            for i in range(n2 - k_nz, n2):
                x[support[i]] = eV[support[i]] * sk.coeff[support[i]].coeff
            y = x * Ht
            eU = vector(F, x[:-1]) # ignore last coordinate of x
            for i in range(n2 - kU):
                eU[support[i]] = -y[i]
            e = sk.phiUV(eU, eV)
            # now sU = eU * HUt
            # print (sU - eU * sk.HUt).hamming_weight(),e.hamming_weight()
            # print e.hamming_weight()
            if (e.hamming_weight() == w):
                break
        if acceptU(t, ellUV(e), rU):
            break
    return e

##############
# DecodeUV
##############

def sign(s, sk, DV, DU, rV, rU, w, delta, *args):
    # a given syndrome will always be decoded with the same ramdomness
    random.seed(str(s))
    n2 = sk.HUt.nrows()
    kU = n2 - sk.HUt.ncols()
    kV = n2 - sk.HVt.ncols()
    n = 2 * n2
    k = kU + kV
    F = sk.HUt.base_ring()
    sprime = s * sk.Sti
    sU = vector([sprime[i] for i in range(n2 - kU)])
    sV = vector([sprime[i] for i in range(n2 - kU, n - k)])
    eV = decodeV(sV, sk, DV, rV, delta)
    eprime = decodeU(sU, sk, eV, DU, rU, w, delta)
    e = vector([eprime[sk.perm[i]] for i in range(n)])
    # if the extra argument 'noperm' is given we return the unpermuted
    # error pattern (for testing purpose)
    if args:
        if args[0] == 'noperm':
            return eprime
    return e
